"""
*****************************************************
*              Filter Pruning
*
*              Gabor Vecsei
* Website:     https://gaborvecsei.com
* LinkedIn:    https://www.linkedin.com/in/gaborvecsei
* Twitter:     https://twitter.com/GAwesomeBE
* Github:      https://github.com/gaborvecsei
* Blog:        https://gaborvecsei.wordpress.com/
*
*****************************************************
"""

from ridurre.base_filter_pruning import BasePruning
from ridurre.kmeans_pruning import KMeansFilterPruning
from ridurre.random_pruning import RandomFilterPruning
from ridurre.ultimate_prune import UltimatePruning

__version__ = "0.0.2"
__author__ = "Gabor Vecsei"
